#ifndef PKMNLIB_ANGELSCRIPTRESOLVER_HPP
#define PKMNLIB_ANGELSCRIPTRESOLVER_HPP

#include <CreatureLib/Battling/ScriptHandling/ScriptResolver.hpp>
#include "../../../extern/angelscript_addons/scriptbuilder/scriptbuilder.h"
#include "../../Battling/Library/BattleLibrary.hpp"
#include "AngelScriptEvolutionScript.hpp"

#define ANGELSCRIPT_DLL_LIBRARY_IMPORT
#include <Arbutils/StringView.hpp>
#include <angelscript.h>
#include "AngelScriptItemUseScript.hpp"
#include "AngelScriptScript.hpp"
#include "AngelScriptTypeInfo.hpp"

class AngelScriptResolver final : public PkmnLib::Battling::ScriptResolver {
private:
    asIScriptEngine* _engine = nullptr;
    asIScriptModule* _mainModule = nullptr;
    ContextPool* _contextPool = nullptr;
    CScriptBuilder _builder = {};
    std::string _sourceDirectory = {};

    ArbUt::Dictionary<ScriptCategory, ArbUt::Dictionary<ArbUt::StringView, AngelScriptTypeInfo*>> _typeDatabase;
    ArbUt::Dictionary<ArbUt::StringView, asITypeInfo*> _baseTypes;
    ArbUt::Dictionary<ArbUt::StringView, asITypeInfo*> _itemUseTypes;
    ArbUt::Dictionary<ArbUt::StringView, asITypeInfo*> _evolutionTypes;
    ArbUt::Dictionary<const CreatureLib::Library::Item*, AngelScriptItemUseScript*> _itemUseScripts;
    ArbUt::Dictionary<ArbUt::StringView, AngelScriptEvolutionScript*> _evolutionScripts;

    static void MessageCallback(const asSMessageInfo* msg, void* param);
    static void Print(const std::string& str) { std::cout << str << std::endl; }
    static void PrintConst(const ArbUt::StringView& str) { Print(std::string(str.std_str())); }
    static i32 IncludeCallback(const char* include, const char* from, CScriptBuilder* builder, void* userParam);

    void RegisterTypes();
    void InitializeByteCode(const ArbUt::Dictionary<i16, ArbUt::Dictionary<ArbUt::StringView, uint32_t>>& types);
    void RegisterScriptType(asITypeInfo* typeInfo, const ArbUt::StringView& metadataKind,
                            const ArbUt::StringView& effectName);

public:
    ~AngelScriptResolver() override {
        for (const auto& ius : _itemUseScripts) {
            delete ius.second;
        }
        delete _contextPool;
        for (const auto& category : _typeDatabase) {
            for (const auto& type : category.second) {
                delete type.second;
            }
        }
        _engine->ShutDownAndRelease();
    }

    void Initialize(CreatureLib::Battling::BattleLibrary* library) override { Initialize(library, true); }
    void Initialize(CreatureLib::Battling::BattleLibrary* library, bool includeStandard);
    void CreateScript(const char* name, const char* script);
    inline void SetSourceDirectory(std::string const& path) noexcept { _sourceDirectory = path; }
    const asIScriptModule* GetMainModule() const noexcept { return _mainModule; }

    void FinalizeModule();

    CreatureLib::Battling::BattleScript* LoadScript(ScriptCategory category,
                                                    const ArbUt::StringView& scriptName) override;
    CreatureLib::Battling::ItemUseScript* LoadItemScript(const CreatureLib::Library::Item* item) override;

    ArbUt::OptionalBorrowedPtr<const PkmnLib::Battling::EvolutionScript>
    LoadEvolutionScript(const ArbUt::StringView& view) override;

    void WriteByteCodeToFile(const char* file, bool stripDebugInfo = false);
    void LoadByteCodeFromFile(const char* file);
    uint8_t* WriteByteCodeToMemory(size_t& size, bool stripDebugInfo = false);
    void LoadByteCodeFromMemory(uint8_t*, size_t size);

    const ArbUt::Dictionary<ScriptCategory, ArbUt::Dictionary<ArbUt::StringView, AngelScriptTypeInfo*>>&
    GetTypeDatabase() const noexcept {
        return _typeDatabase;
    }

    void RegisterType(const char* type) {
        int r = _engine->RegisterObjectType(type, 0, asOBJ_REF | asOBJ_NOCOUNT);
        Ensure(r >= 0);
    }
    void RegisterTypeMethod(const char* type, const char* decl, void*(func)(void*)) {
        int r = _engine->RegisterObjectMethod(type, decl, asFunctionPtr(func), asCALL_CDECL_OBJFIRST);
        Ensure(r >= 0);
    }
    void RegisterGlobalMethod(const char*, void*(func)(void*)) {
        auto r = _engine->RegisterGlobalFunction("decl", asFunctionPtr(func), asCALL_CDECL);
        Ensure(r >= 0);
    }

    asITypeInfo* GetBaseType(const ArbUt::StringView& name) {
        asITypeInfo* t;
        auto v = _baseTypes.TryGet(name);
        if (!v.has_value()) {
            t = _mainModule->GetTypeInfoByName(name.c_str());
            if (t == NULL) {
                t = _engine->GetTypeInfoByDecl(name.c_str());
            }
            _baseTypes.Insert(name, t);
        } else {
            t = v.value();
        }
        return t;
    }

    inline ContextPool* GetContextPool() const noexcept { return _contextPool; }
};
#endif // PKMNLIB_ANGELSCRIPRESOLVER_HPP
